#
# ident	"@(#)check_host.tcl	1.38	98/03/05 SMI"
#
# Copyright (c) 1996 by Sun Microsystems, Inc.
# All rights reserved.
#
# check_host is a SSP command that determines whether the host, specified
# by the HOSTNAME env variable, is UP or DOWN.
# The host is considered UP, if all configured procs are running the O/S
# and the signature_block state is RUN. Otherwise, the host is considered
# DOWN.
# The signature block is read from the MIB, kept by the snmpd daemon.
#
# Usage: (Note: not all cmd line args are advertised)
#
#	check_host [-q] [-v] [-b] [-B] [-H] [-p proc] [-h hostname]
#	-q : quiet mode
#	-v : verbose mode
#	-b : reserved for bringup
#	-B : reserved for bringup. Log sigs in the domain messages file.
#	-H : check heartbeat. Use with -b
#	-p proc : check only signatures on proc
#	-h hostname : treat hostname as SUNW_HOSTNAME
#
#	0 - Host is UP
#	1 - Host is Down
#

proc check_host args \
{
	global env
	set usage "usage: check_host \[-q\]"
	set bHostDown 0

	#
	# Good OS states
	#
	set OS(run) 1
	set OS(detached) 1
	set OS(callback) 1
	set OS(offline) 1
	set OS(quiesced) 1
	set OS(quiesce-in-progress) 1
	set OS(resume-in-progress) 1

	#
	# Good OBP states
	#
	set OBP(callback) 1

	#
	# Check command line arguments
	#
	set validArgs "-qbBvHp:h:"

	if { [ catch {CLA_Parse Args $args $validArgs} errMsg ] } {
		puts stderr "Error: $errMsg"
		puts stderr $usage
		return -1
	}

	if { [info exists Args(q)] } {
		# Don't output any msg
		set quiet 1
	} else {
		set quiet 0
	}

	if { [info exists Args(B)] } {
		# verbosity level for bringup
		set bVerbose 1
	} else {
		set bVerbose 0
	}

	if { [info exists Args(v)] } {
		set internal 1
	} else {
		set internal 0
	}

	#
	# Check SSP environment variables:
	#
	set ssp_env "SUNW_HOSTNAME SSPETC SSPVAR"

	foreach var $ssp_env {
		if { ![info exists env($var)] } {
			puts stderr "check_host: Environment variable '$var' must be set."
			return -1
		}
	}

	if { [info exists Args(h)] } {
		set env(SUNW_HOSTNAME) $Args(h)
	}

        #if not exists  p as param, use ping to quickly decide  host is UP.
        #if ping failed, it doesn't necessary mean host is down. it could be 
        # network trafic, go on with more checking. arg p needs further 
        # checking to decide host is UP or DOWN
        if { ![info exists Args(p)] } { 
            if { [catch { exec ping $env(SUNW_HOSTNAME) 3}  errmsg ] == 0  } {
   	        if { $quiet == 0 } { puts "Host is UP" }
                return 0
            }
        }
	#
	# Get a SNMP session handle to talk to the snmpd
	#
	if { [catch {snmp session -community check_host } handle] != 0 } {
		puts stderr "check_host: error creating snmp session - $handle"
		return -1
	}

	if { $internal == 1 } { puts "  snmp session is $handle" }

	# Do we want the sig of just one cpu?
	if { [info exists Args(p)] } {
		return [ check_host_p $handle $Args(p) ]
	}

	#
	# Query snmpd for domain's sysbrd and proc config list
	#
	set proc_list ""
	set brd_list ""
	set validhost 0

	if { [catch { $handle walk x "domainIndex domainName" {
		set indx [lindex [lindex $x 0] 2]
		set host [lindex [lindex $x 1] 2]
		if { $host == $env(SUNW_HOSTNAME) } {
			# Got our domain name
			set brd_list [lindex [lindex [$handle get domainSysBrdConfig.$indx] 0] 2]
			set proc_list [lindex [lindex [$handle get domainProcConfig.$indx] 0] 2]
			set validhost 1
			break
		}
	} } error ] != 0 }  {  
               if { $quiet == 0 } {
                    puts $error
               }
               $handle destroy
               return -1 
        } 


	#
	# Confirm if SUNW_HOSTNAME is set to a valid domain name
	#
	if { $validhost == 0 } {
		if { $quiet == 0 } {
			puts "$env(SUNW_HOSTNAME) is not a valid domain"
		}
		if { $bVerbose == 1 } {
			logMessage "$env(SUNW_HOSTNAME) is not a valid domain"
		}
                $handle destroy
		return 1
	}

	#
	# Check that no sysbrds have arb-stopped
	#
	set brd_list [ split $brd_list - ]
	if { $brd_list == "" } {
		if { $quiet == 0 } { puts "Host is DOWN" }
		if { $bVerbose == 1 } {
			logDomainMessage "No SNMP sysbrd config for this domain" "check_host"
		}
		$handle destroy
		return 1
	}
	if { $internal == 1 } { puts "  System Board list is $brd_list" }
        if { [catch { check_laarb $brd_list}  arb ] == 0  } {
	   set stop_chk [ lindex $arb 1 ] 
  	   if { $stop_chk != 0 } {
		if { $quiet == 0 } { puts "Host is DOWN" }
		if { $bVerbose == 1 } {
			set message [format "WARNING : arb-stop detected on system board %s" \
				[lindex $arb 0]]
			logDomainMessage $message "check_host"
		}
		$handle destroy
		return 1
	   }
        }

	#
	# Continue on with proc's signature and state
	#
	set proc_config_list [ split $proc_list - ]
	if { $proc_config_list == "" } {
		if { $quiet == 0 } { puts "Host is DOWN" }
		if { $bVerbose == 1 } {
			logDomainMessage "No SNMP proc config for this domain" "check_host"
		}
		$handle destroy
		return 1
	}

	if { $internal == 1 } { puts "  Processor config list is $proc_config_list" }

	#
	# Build a good proc list from processors in the proc config list
	# marked good
	#
	set good_proc_list ""
	foreach i $proc_config_list {
		set b_num 0
		set p_num 0
		set p_status "F"
		scan [ split $i . ] "%d %d %s" b_num p_num p_status

		if { $p_status == "G" } {
			set p [expr $b_num * 4 + $p_num]
			lappend good_proc_list $p
		}
	}

	if { $good_proc_list == "" } {
		if { $quiet == 0 } { puts "Host is DOWN" }
		if { $bVerbose == 1 } {
			logDomainMessage "No SNMP good proc config for this domain" "check_host"
		}
		$handle destroy
		return 1
	}

	if { $internal == 1 } { puts "  Good processor config list is $good_proc_list" }

	#
	# Convert good proc list to good index list
	#
	set good_index_list ""
	set PROC_STATE_INDEXES "procStateGenNum procStateGenIndex"
	if { [ catch {$handle walk x  "$PROC_STATE_INDEXES"   {
	    set num [lindex [lindex $x 0] 2]
	    if { [ lsearch $good_proc_list $num ] >= 0 } {
			# This proc belongs to this domain
			set index [lindex [lindex $x 1] 2]
			lappend good_index_list $index
	    }
        } } error ] == 0 }  {
	    if { $good_index_list == "" } {
	        if { $quiet == 0 } { puts "Host is DOWN" }
  	        if { $bVerbose == 1 } {
	    	    logDomainMessage "No SNMP good proc config index for this domain" "check_host"
	        }
	        $handle destroy
	        return 1
            }
        }
        

	if { $internal == 1 } { puts "  Good indexes are: $good_index_list" }

	if { [info exists Args(b)] } {
                # with empty good_index_list because reading MIB failure, 
                # check_host_B returns UP
		return [ check_host_B $handle $good_index_list ]
	}

	#
	# Read the signature block (state and sig fields) for each proc
	# in good_index_list.
	#
	foreach index $good_index_list {
		if { [catch {$handle get "procStateGenPgmSignature.$index procStateGenPgmState.$index" } data] != 0 } {
			# safer to assume Host is UP
			break
		}
		set sig   [lindex [lindex $data 0] 2]
		set state [lindex [lindex $data 1] 2]

		if { $internal == 1 } {
			puts "  Processor index $index signature is $sig, state is $state"
		}

		if { ($sig == "os") && ([info exists OS($state)]) } {
			continue
		} elseif { ($sig == "obp") && ([info exists OBP($state)]) } {
			continue
		} else {
			# Host is down
			incr bHostDown
			break
		}
	}
	if { $quiet == 0 } {
		if { $bHostDown } {
			puts "Host is DOWN"
			dump_signatures $handle $good_index_list $bVerbose
			$handle destroy
			return 1
		} else {
			puts "Host is UP"
			$handle destroy
			return 0
		}
	}

	if { $bHostDown } {
		dump_signatures $handle $good_index_list $bVerbose
		$handle destroy
		return 1
	}
	$handle destroy
	return 0
}


#
# check_host_B: this proc is used exclusively by bringup to decide
# whether or not to use the -C argument to configure the centerplane.
# This proc checks state, signature, and heartbeat of each processor
# before determinig if the Host is UP or DOWN
# with an empty index_list, check_host_B returns UP
#
proc check_host_B { SnmpSession index_list } \
{
	upvar quiet bQuiet
	upvar bVerbose Verbose
	upvar internal Internal
	upvar Args Arrgg
        set sig ""

	foreach i $index_list {
		if { [catch {$SnmpSession get "procStateGenNum.$i procStateGenPgmSignature.$i \
				procStateGenPgmState.$i procStateGenPgmSubState.$i"} data] != 0 } {
			# Host is down ? it is safer to assume UP
			puts "Warning: cannot determine host state and sig, Index $i"
			puts "Error: $data"
			dump_signatures $SnmpSession $index_list $Verbose
			$SnmpSession destroy
			return 0
		}
		set proc [lindex [lindex $data 0] 2]
		set sig [lindex [lindex $data 1] 2]
		set state [lindex [lindex $data 2] 2]
		set subState [lindex [lindex $data 3] 2]

		if { $Internal == 1 } {
			puts "proc $proc, sig $sig, state $state, subState $subState"
		}

		switch $sig {
			post {
				# Don't care. Note: bringup checks for hpost process
				if { $bQuiet == 0 } { puts "Host is DOWN, $sig $state" }
				dump_signatures $SnmpSession $index_list $Verbose
				$SnmpSession destroy
				return 1
			}

			download-helper {
				if { $state == "run" } {
					# Note: need validate just one proc
					if { $bQuiet == 0 } { puts "Host is ALIVE, $sig running" }
					$SnmpSession destroy
					return 0
				} else {
					# State is either unknown or exit
					if { $bQuiet == 0 } { puts "Host is DOWN, $sig $state" }
					dump_signatures $SnmpSession $index_list $Verbose
					$SnmpSession destroy
					return 1
				}
			}

			obp {
				# Check state and subState. Note: need to validate just one proc
				if { ($state == "run") || ($state == "callback") } {
					if { $bQuiet == 0 } { puts "Host is ALIVE, $sig $state" }
					$SnmpSession destroy
					return 0
				}

				# State not running
				if { (($state == "exit") &&
					(($subState == "exit-reboot") ||
					 ($subState == "exit-panic-reboot"))) } {
					# Reboot in progress
					if { $bQuiet == 0 } { puts "$sig $state $subState. Reboot in progress" }
					$SnmpSession destroy
					return 0
				} elseif { $state == "booting" } {
					# Boot in progress
					if { $bQuiet == 0 } { puts "$sig $state $subState. OBP booting" }
					$SnmpSession destroy
					return 0
				} elseif { ($state == "exit") &&
						(($subState == "exit-halt") || ($subState == "exit-obp-reset")) } {
					# OS halted, may reboot
					if { $bQuiet == 0 } { puts "Host is ACTIVE, $sig $state $subState" }
					$SnmpSession destroy
					return 0
				} elseif { ($state == "watchdog") || ($state == "watchdog-sync") } {
					if { $bQuiet == 0 } { puts "Host is ACTIVE, $sig $state $subState" }
					$SnmpSession destroy
					return 0
				} elseif { ($state == "redmode") || ($state == "redmode-sync") } {
					if { $bQuiet == 0 } { puts "Host is ACTIVE, $sig $state $subState" }
					$SnmpSession destroy
					return 0
				} else {
					# Bad OBP state
					if { $bQuiet == 0 } { puts "Host is DOWN, $sig $state $subState" }
					dump_signatures $SnmpSession $index_list $Verbose
					$SnmpSession destroy
					return 1
				}
			}

			os {
				if { $state == "run" } {
					if { [info exists Arrgg(H)] } {
					# Check heartbeat. Note: arb-stops already checked
					set ohb ""
					for { set j 1 } { $j < 4 } { incr j } {
						if { [catch {$SnmpSession get procStateGenHeartbeat.$i} hb ] != 0 } {
							# Host is down ?
							puts "Warning: cannot get heartbeat from procStateGenHeartbeat.$i"
							puts Error: $hb"
						} else {
							set hb [lindex [lindex $hb 0] 2]
							if { $Internal == 1 } {
								puts "proc $proc, sig $sig, state $state, subState $subState hb $hb"
							}
							if { $ohb == $hb } {
								# OS not beating. Host is DOWN
								if { $bQuiet == 0 } { puts "Host is DOWN" }
								$SnmpSession destroy
								return 1
							} else {
								set ohb $hb
							}
						}
						# time delay
						sleep 3
					}
					}
				} elseif { ($state == "watchdog-sync") ||
							($state == "xir-sync") ||
							($state == "detached") ||
							($state == "callback") ||
							($state == "offline") ||
							($state == "quiesced") ||
							($state == "quiesce-in-progress") ||
							($state == "resume-in-progress") } {
					# domain is active, continue
				} elseif { (($state == "exit") &&
							(($subState == "exit-panic1") ||
							 ($subState == "exit-panic2") ||
							 ($subState == "exit-reboot") ||
							 ($subState == "exit-panic-reboot"))) } {
					# OS panic'ed and edd will reboot
					if { $bQuiet == 0 } {
						puts "Host is ACTIVE, $sig $state $subState, reboot in progress"
					}
					$SnmpSession destroy
					return 0
				} else {
					# OS hang! Host is Down
					if { $bQuiet == 0 } { puts "Host is DOWN, $sig $state $subState" }
					dump_signatures $SnmpSession $index_list $Verbose
					$SnmpSession destroy
					return 1
				}
			}

			default {
				# unknown sig
				if { $Internal == 1 } {
					puts "Error: unknown signature `$sig` on proc $proc"
				}
				if { $bQuiet == 0 } { puts "Host is DOWN, $sig $state $subState" }
				dump_signatures $SnmpSession $index_list $Verbose
				$SnmpSession destroy
				return 1
			}
		}
	}

	# Host is either UP or coming UP
	if { $bQuiet == 0 } {
		if { $sig == "os" } {
			puts "Host is UP"
		} else {
			puts "Host is ALIVE, $sig running"
		}
	}
	$SnmpSession destroy
	return 0
}


#
# check_host_p: this procedure is used exclusively by bringup to check
# signatures on this processor only.
# when reading MIB failed, check_host_p returns DOWN
#
proc check_host_p { SnmpSession cpu } \
{
	upvar bVerbose Verbose
	set sig ""
	set state ""

	if { $cpu < 0 || $cpu >= 64 } {
		if { $Verbose == 1 } {
			set message [format "invalid processor number: %d" $cpu]
			logDomainMessage $message "check_host"
		}
		$SnmpSession destroy
		return -1
	}

	set PROC_STATE_TBL "procStateGenNum procStateGenPgmSignature procStateGenPgmState"
	catch {$SnmpSession walk x  "$PROC_STATE_TBL"  {
		set proc [lindex [lindex $x 0] 2]
		if { $proc == $cpu } {
			# Found proc. Check signatures
			set sig [lindex [lindex $x 1] 2]
			set state [lindex [lindex $x 2] 2]
			break
		}
	} } error 

        if { ($sig == "") || ($state == "") } {
		if { $Verbose == 1 } {
			set message [format "signatures on proc %d are empty" $cpu ]
			logDomainMessage $message "check_host"
		}
		set res 1
        } else {
		if { (($state == "run") &&
			(($sig == "download-helper") || ($sig == "obp") || ($sig == "os"))) } {
			set res 0
		} else {
			set res 1
		}
        }

	$SnmpSession destroy
	if { ($res == 1) && ($Verbose == 1) } {
		set message [format "proc %s: sig %s and state %s" $proc $sig $state]
		logDomainMessage $message "check_host"
	}
	return $res
}


#
# This proc is added for bringup -B, so it can tell us to log signatures
# into the domain messages file. NOTE: we break up signatures in small
# chunks, otherwise logger core dumps.
#
proc dump_signatures { SnmpSession index verbose } \
{
	set message {}
	set j 0

	if { $verbose == 0 } { return }

	foreach i $index {
		if { [ catch {$SnmpSession get "procStateGenNum.$i procStateGenPgmSignature.$i \
				procStateGenPgmState.$i procStateGenPgmSubState.$i"} x ] == 0 } {
		    append message [format "%d %s %s %s, " \
			[lindex [lindex $x 0] 2] [lindex [lindex $x 1] 2] \
			[lindex [lindex $x 2] 2] [lindex [lindex $x 3] 2] ]
		    incr j
		    if { $j == 10 } {
			logDomainMessage $message "check_host"
			#puts "check_host $message"
			set j 0
			set message {}
		    }
                }
	}
	logDomainMessage $message "check_host"
	#puts "check_host $message"
}
